package ch.unige.aou.service.storage;

import java.io.FileOutputStream;
import java.io.IOException;
import java.net.URI;
import java.nio.file.Path;
import java.nio.file.Paths;

import org.springframework.boot.autoconfigure.condition.ConditionalOnBean;
import org.springframework.context.annotation.Profile;
import org.springframework.stereotype.Service;

import ch.unige.solidify.SolidifyConstants;
import ch.unige.solidify.exception.SolidifyProcessingException;
import ch.unige.solidify.exception.SolidifyRuntimeException;
import ch.unige.solidify.service.MessageService;
import ch.unige.solidify.util.FileTool;

import ch.unige.aou.AouConstants;
import ch.unige.aou.business.PublicationService;
import ch.unige.aou.business.UserService;
import ch.unige.aou.config.AouProperties;
import ch.unige.aou.controller.AdminController;
import ch.unige.aou.model.publication.DocumentFile;
import ch.unige.aou.model.publication.Publication;
import ch.unige.aou.service.MetadataService;

@Profile("storage-file")
@Service
@ConditionalOnBean(AdminController.class)
public class FileStorageService extends StorageService {

  private final Path storagePath;

  public FileStorageService(AouProperties aouProperties, MessageService messageService, PublicationService publicationService,
          UserService userService, MetadataService metadataService) {
    super(messageService, publicationService, userService, metadataService);
    
    // Check configuration
    this.storagePath = Paths.get(URI.create(aouProperties.getStorage().getUrl()));
    if (!FileTool.checkFile(this.storagePath)) {
      throw new SolidifyRuntimeException("Storage location (" + this.storagePath + ") does not exist");
    }
  }

  @Override
  public void generateExportMetadata(Publication publication) {
    // Do nothing in this implementation
  }

  @Override
  public void storePublication(Publication publication) throws IOException {
    // Create publication folder
    Path publicationFolder = this.storagePath.resolve(publication.getResId());
    if (!FileTool.checkFile(publicationFolder) && !FileTool.ensureFolderExists(publicationFolder)) {
      throw new SolidifyProcessingException("Cannot create publication folder " + publication.toString());
    }
    // Save location
    publication.setArchiveId(publicationFolder.toUri().toString());
    // Save metadata
    try (FileOutputStream aouMetadata = new FileOutputStream(
            publicationFolder.resolve(AouConstants.AOU + SolidifyConstants.XML_EXT).toFile())) {
      aouMetadata.write(publication.getMetadata().getBytes());
    }
    // Save attachments
    for (DocumentFile attachment : publication.getDocumentFiles()) {
      if (!FileTool.copyFile(Paths.get(attachment.getFinalData()), publicationFolder.resolve(attachment.getFileName()))) {
        throw new SolidifyProcessingException("Cannot save publication attachment " + attachment.getFinalData().toString());
      }
    }
  }

}
