package ch.unige.aou.service;

import java.lang.reflect.Array;
import java.net.URI;
import java.nio.file.Path;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.core.env.Environment;
import org.springframework.http.HttpEntity;
import org.springframework.http.HttpMethod;
import org.springframework.web.client.RestTemplate;

import ch.unige.solidify.SolidifyConstants;
import ch.unige.solidify.rest.ActionName;
import ch.unige.solidify.rest.CollectionPage;
import ch.unige.solidify.rest.Resource;
import ch.unige.solidify.security.TokenUsage;
import ch.unige.solidify.util.AbstractRestClientTool;
import ch.unige.solidify.util.CollectionTool;

public abstract class ResourceClientService<T extends Resource<T>> extends ResourceBaseClientService<T> implements
        ResourceClientServiceInterface<T> {
  private static final int BUFFER_SIZE = 16384; // 2 ^ 14
  private static final Logger logger = LoggerFactory.getLogger(ResourceClientService.class);

  public ResourceClientService(String resourcePath, Class<T> resourceClass, String module, AbstractRestClientTool restClientTool,
          Environment env) {
    super(resourcePath, resourceClass, module, restClientTool, env);
  }

  public ResourceClientService(String resourcePath, Class<T> resourceClass, String module, AbstractRestClientTool restClientTool,
          Environment env,
          boolean isResourceServer) {
    super(resourcePath, resourceClass, module, restClientTool, env);
    this.setResourceServer(isResourceServer);
  }

  @Override
  public T create(T resource) {
    final String url = this.getBaseUrl() + this.resourcePath;
    final RestTemplate restTemplate = this.restClientTool.getClient();
    return restTemplate.postForObject(url, resource, this.getResourceClass());
  }

  @Override
  public void delete(String resId) {
    final String url = this.getBaseUrl() + this.resourcePath + SolidifyConstants.URL_ID;
    final RestTemplate restTemplate = this.restClientTool.getClient();
    final Map<String, String> params = new HashMap<>();
    params.put(SolidifyConstants.URL_ID_FIELD, resId);
    restTemplate.delete(url, params);
  }

  @Override
  public void deleteList(String[] ids) {
    final String url = this.getBaseUrl() + this.resourcePath;
    final RestTemplate restTemplate = this.restClientTool.getClient();
    restTemplate.exchange(url, HttpMethod.DELETE, new HttpEntity<>(ids), String.class);
  }

  @Override
  public T findOne(String resId) {
    final String url = this.getBaseUrl() + this.resourcePath + SolidifyConstants.URL_ID;
    final Map<String, String> params = new HashMap<>();
    params.put(SolidifyConstants.URL_ID_FIELD, resId);
    return this.getObjectFromUrl(url, params);
  }

  @Override
  public T update(String resId, T resource) {
    final String url = this.getBaseUrl() + this.resourcePath + SolidifyConstants.URL_ID;
    final RestTemplate restTemplate = this.restClientTool.getClient();
    final Map<String, String> params = new HashMap<>();
    params.put(SolidifyConstants.URL_ID_FIELD, resId);
    return restTemplate.exchange(url, HttpMethod.PATCH, new HttpEntity<>(resource), this.getResourceClass(), params).getBody();
  }

  protected <U extends Resource<U>> void addItem(String resId, String itemTypeLabel, Class<U> itemClass, String... ids) {
    final Map<String, String> params = new HashMap<>();
    params.put(SolidifyConstants.URL_ID_FIELD, resId);
    final RestTemplate restTemplate = this.restClientTool.getClient();
    restTemplate.postForObject(this.getAssociationUrl(itemTypeLabel), ids, itemClass, params);
  }

  protected <U extends Resource<U>> void addItem(String resId, String itemTypeLabel, String itemId, Class<U> itemClass) {
    @SuppressWarnings("unchecked")
    final Class<? extends Resource<U>[]> arrayType = (Class<? extends Resource<U>[]>) Array.newInstance(itemClass, 0).getClass();
    final Map<String, String> params = new HashMap<>();
    params.put(SolidifyConstants.URL_ID_FIELD, resId);
    final RestTemplate restTemplate = this.restClientTool.getClient();
    restTemplate.postForObject(this.getAssociationUrl(itemTypeLabel), new String[] { itemId }, arrayType, params);
  }

  protected <U extends Resource<U>> U addItem(String resId, String itemTypeLabel, U item, Class<U> itemClass) {
    final RestTemplate restTemplate = this.restClientTool.getClient();
    return restTemplate.postForObject(this.getAssociationUrl(itemTypeLabel), item, itemClass, resId);
  }

  protected <U extends Resource<U>> void addRelationItem(String resId, String itemTypeLabel, Class<U> itemClass, String... ids) {
    @SuppressWarnings("unchecked")
    final Class<? extends Resource<U>[]> arrayType = (Class<? extends Resource<U>[]>) Array.newInstance(itemClass, 0).getClass();
    final Map<String, String> params = new HashMap<>();
    params.put(SolidifyConstants.URL_ID_FIELD, resId);
    final RestTemplate restTemplate = this.restClientTool.getClient();
    restTemplate.postForObject(this.getAssociationUrl(itemTypeLabel), ids, arrayType, params);
  }

  protected T doAction(String action) {
    final String url = this.getActionUrl(action);
    return this.getObjectFromUrl(url, new HashMap<>());
  }

  protected List<T> doActionOnList(String action, Class<T> itemClass) {
    final String url = this.getActionUrl(action) + "?size=" + CollectionPage.MAX_SIZE_PAGE;
    final RestTemplate restTemplate = this.restClientTool.getClient();
    final String jsonString = restTemplate.getForObject(url, String.class);
    return CollectionTool.getList(jsonString, itemClass);
  }

  protected List<T> docActionOnListWithParameter(String action, Class<T> itemClass, String parameterName, String parameterValue) {
    final String url = this.getActionUrl(action) + "?size=" + CollectionPage.MAX_SIZE_PAGE + "&" + parameterName + "=" + parameterValue;
    final RestTemplate restTemplate = this.restClientTool.getClient();
    final String jsonString = restTemplate.getForObject(url, String.class);
    return CollectionTool.getList(jsonString, itemClass);
  }

  protected T doActionOnId(String resId, String action) {
    final String url = this.getAssociationUrl(action);
    final Map<String, String> params = new HashMap<>();
    params.put(SolidifyConstants.URL_ID_FIELD, resId);
    return this.getObjectFromUrl(url, params);
  }

  protected T doActionWithId(String resId, String action) {
    final String url = this.getActionUrlWithId(action);
    final Map<String, String> params = new HashMap<>();
    params.put(SolidifyConstants.URL_ID_FIELD, resId);
    return this.getObjectFromUrl(url, params);
  }

  protected void downloadByUrlWithToken(String urlString, Path path) {
    final String url = this.getBaseUrl() + urlString;
    this.restClientTool.downloadContent(URI.create(url), path, TokenUsage.WITH_TOKEN);
  }

  protected void downloadDeposit(String resId, String folder, Path path) {
    String url = null;
    if (folder == null) {
      url = this.getBaseUrl() + this.resourcePath + SolidifyConstants.URL_SEP + resId + SolidifyConstants.URL_SEP + ActionName.DOWNLOAD;
    } else {
      url = this.getBaseUrl() + this.resourcePath + SolidifyConstants.URL_SEP + resId + SolidifyConstants.URL_SEP + ActionName.DOWNLOAD
              + "?folder=" + folder;
    }
    this.restClientTool.downloadContent(URI.create(url), path, TokenUsage.WITH_TOKEN);
  }

  protected List<T> findAll(Class<T> itemClass) {
    final String url = this.getBaseUrl() + this.resourcePath + "?size=" + CollectionPage.MAX_SIZE_PAGE;
    final RestTemplate restTemplate = this.restClientTool.getClient();
    final String jsonString = restTemplate.getForObject(url, String.class);
    return CollectionTool.getList(jsonString, itemClass);
  }

  protected <U> List<U> findAllLinkedResources(String resId, String itemTypeLabel, Class<U> itemClass) {
    final String url = this.getBaseUrl() + this.resourcePath + SolidifyConstants.URL_ID + SolidifyConstants.URL_SEP + itemTypeLabel
            + "?size=" + CollectionPage.MAX_SIZE_PAGE;
    final Map<String, String> params = new HashMap<>();
    params.put(SolidifyConstants.URL_ID_FIELD, resId);
    final RestTemplate restTemplate = this.restClientTool.getClient();
    final String jsonString = restTemplate.getForObject(url, String.class, params);
    return CollectionTool.getList(jsonString, itemClass);
  }

  protected <U> U getChildItemFromRelation(String resId, String itemTypeLabel, Class<U> itemClass, String childId) {
    final String url = this.getBaseUrl() + this.resourcePath + SolidifyConstants.URL_ID + SolidifyConstants.URL_SEP + itemTypeLabel
            + SolidifyConstants.URL_SEP + childId;
    final Map<String, String> params = new HashMap<>();
    params.put(SolidifyConstants.URL_ID_FIELD, resId);
    final RestTemplate restTemplate = this.restClientTool.getClient();
    return restTemplate.getForObject(url, itemClass, params);
  }

  protected void removeItem(String parentId, String itemTypeLabel, String itemId) {
    final String url = this.getBaseUrl() + this.resourcePath + SolidifyConstants.URL_SEP + parentId + SolidifyConstants.URL_SEP + itemTypeLabel
            + SolidifyConstants.URL_SEP + itemId;
    final RestTemplate restTemplate = this.restClientTool.getClient();
    restTemplate.delete(url);
  }

  protected void removeItem(String parentId, String itemTypeLabel, String itemId, String... subItemIds) {
    if (subItemIds.length == 0) {
      this.removeItem(parentId, itemTypeLabel, itemId);
    } else {
      final Map<String, String> params = new HashMap<>();
      params.put(SolidifyConstants.URL_ID_FIELD, parentId);
      final RestTemplate restTemplate = this.restClientTool.getClient();
      restTemplate
              .exchange(this.getAssociationUrl(itemTypeLabel + SolidifyConstants.URL_SEP + itemId), HttpMethod.DELETE,
                      new HttpEntity<>(subItemIds),
                      String.class, params);
    }
  }

  protected List<T> searchByProperties(Map<String, String> params, Class<T> itemClass) {
    String url = this.getBaseUrl() + this.resourcePath + "?size=" + CollectionPage.MAX_SIZE_PAGE;
    for (final Map.Entry<String, String> entry : params.entrySet()) {
      url += "&" + entry.getKey() + "=" + entry.getValue();
    }
    final RestTemplate restTemplate = this.restClientTool.getClient();
    final String jsonString = restTemplate.getForObject(url, String.class);
    return CollectionTool.getList(jsonString, itemClass);
  }

  protected <U extends Resource<U>> U updateSubItem(String resId, String subItemTypeLabel, Class<U> subItemClass, String subItemId,
          Map<String, Object> mapProperties) {
    final Map<String, String> params = new HashMap<>();
    params.put(SolidifyConstants.URL_ID_FIELD, resId);
    final RestTemplate restTemplate = this.restClientTool.getClient();
    return restTemplate.exchange(
            this.getAssociationUrl(subItemTypeLabel) + SolidifyConstants.URL_SEP + subItemId,
            HttpMethod.PATCH,
            new HttpEntity<>(mapProperties),
            subItemClass,
            params)
            .getBody();
  }

  private String getActionUrl(String actionLabel) {
    return this.getBaseUrl() + this.resourcePath + SolidifyConstants.URL_SEP + actionLabel;
  }

  private String getActionUrlWithId(String actionLabel) {
    return this.getBaseUrl() + this.resourcePath + SolidifyConstants.URL_SEP + actionLabel + SolidifyConstants.URL_SEP + "{"
            + SolidifyConstants.URL_ID_FIELD
            + "}";
  }

  protected String getAssociationUrl(String resourceLabel) {
    return this.getBaseUrl() + this.resourcePath + SolidifyConstants.URL_ID + SolidifyConstants.URL_SEP + resourceLabel;
  }

}
