package ch.unige.aou.business;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.stereotype.Service;

import ch.unige.aou.AouConstants;
import ch.unige.aou.model.settings.Label;
import ch.unige.aou.model.settings.Language;
import ch.unige.aou.model.settings.ValueWithTranslations;
import ch.unige.aou.repository.LanguageRepository;
import ch.unige.aou.specification.LanguageSpecification;

@Service
public class LanguageService extends AouResourceService<Language> {
  static final Logger log = LoggerFactory.getLogger(LanguageService.class);

  private LanguageRepository languageRepository;

  public LanguageService(LanguageRepository languageRepository) {
    this.languageRepository = languageRepository;
  }

  public Language findByCode(String code) {
    return this.languageRepository.findByCode(code);
  }

  @Override
  public LanguageSpecification getSpecification(Language resource) {
    return new LanguageSpecification(resource);
  }

  public void initDefaultData() {

    /**
     * Default languages using ISO 639-2 code which is the format devised for use by libraries, information services, and publishers
     * to indicate language in the exchange of information, especially in computerized systems.
     * See https://www.loc.gov/standards/iso639-2/faq.html#5
     */
    List<ValueWithTranslations> valueWithTranslations = Arrays.asList(
            this.newValueWithEngFreTranslations("fre", "French", "Français"),
            this.newValueWithEngFreTranslations("eng", "English", "Anglais"),
            this.newValueWithEngFreTranslations("ger", "German", "Allemand"),
            this.newValueWithEngFreTranslations("ara", "Arabic", "Arabe"),
            this.newValueWithEngFreTranslations("chi", "Chinese", "Chinois"),
            this.newValueWithEngFreTranslations("spa", "Spanish", "Espagnol"),
            this.newValueWithEngFreTranslations("gre", "Greek", "Grec"),
            this.newValueWithEngFreTranslations("ita", "Italian", "Italien"),
            this.newValueWithEngFreTranslations("jpn", "Japanese", "Japonais"),
            this.newValueWithEngFreTranslations("dut", "Dutch", "Néerlandais"),
            this.newValueWithEngFreTranslations("por", "Portuguese", "Portugais"),
            this.newValueWithEngFreTranslations("roh", "Romansh", "Romanche"),
            this.newValueWithEngFreTranslations("rus", "Russian", "Russe"),
            this.newValueWithEngFreTranslations("tur", "Turkish", "Turc"),
            this.newValueWithEngFreTranslations("mis", "Other", "Autre")
    );

    /**
     * First ensure all languages exist without labels
     */
    List<String> newLanguageCodes = new ArrayList<>();
    int sortValue = 0;
    for (ValueWithTranslations valueWithTranslation : valueWithTranslations) {
      if (this.languageRepository.findByCode(valueWithTranslation.getValue()) == null) {
        sortValue = sortValue + AouConstants.ORDER_INCREMENT;
        Language language = new Language();
        language.setResId(valueWithTranslation.getValue().toUpperCase());
        language.setCode(valueWithTranslation.getValue());
        language.setSortValue(sortValue);
        this.save(language);
        log.info("Language '{}' created", valueWithTranslation.getValue());
        newLanguageCodes.add(valueWithTranslation.getValue());
      }
    }

    /**
     * Only then check if their labels exist, as label's languages must already exist to be created
     */
    for (ValueWithTranslations valueWithTranslation : valueWithTranslations) {
      if (newLanguageCodes.contains(valueWithTranslation.getValue())) {
        Language language = this.findByCode(valueWithTranslation.getValue());
        List<Label> labels = new ArrayList<>();
        labels.add(
                new Label(valueWithTranslation.getTranslation(AouConstants.LANG_CODE_FRENCH), this.findByCode(AouConstants.LANG_CODE_FRENCH)));
        labels.add(
                new Label(valueWithTranslation.getTranslation(AouConstants.LANG_CODE_ENGLISH), this.findByCode(AouConstants.LANG_CODE_ENGLISH)));
        language.setLabels(labels);
        this.save(language);
        log.info("Language's labels created for language '{}'", valueWithTranslation.getValue());
      }
    }
  }
}
