package ch.unige.aou.service.metadata.imports;

import java.io.IOException;
import java.security.NoSuchAlgorithmException;
import java.util.HashMap;
import java.util.Map;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.http.HttpStatus;
import org.springframework.stereotype.Service;

import ch.unige.solidify.exception.SolidifyHttpErrorException;
import ch.unige.solidify.exception.SolidifyRuntimeException;
import ch.unige.solidify.service.MessageService;
import ch.unige.solidify.util.StringTool;

import ch.unige.aou.config.AouProperties;
import ch.unige.aou.exception.AouRuntimeException;
import ch.unige.aou.exception.ErrorCodes;
import ch.unige.aou.model.ExistingPublicationInfo;
import ch.unige.aou.model.publication.Publication;
import ch.unige.aou.model.tool.ValidationTool;
import ch.unige.aou.model.xml.deposit.v2_4.DepositDoc;
import ch.unige.aou.service.MetadataService;
import ch.unige.aou.service.metadata.MetadataExtractor;
import ch.unige.aou.service.rest.DuplicateService;

@Service
public class MetadataImportWorkFlowService {
  private static final Logger log = LoggerFactory.getLogger(MetadataImportWorkFlowService.class);

  private DoiCrossrefImportService doiCrossrefImportService;
  private ArxivImportService arxivImportService;
  private PmidImportService pmidImportService;
  private InspireHepImportService inspireHepImportService;
  private EuropePmcImportService europePmcImportService;
  private DataCiteImportService dataCiteImportService;
  private final UnpaywallImportService unpaywallImportService;
  private final MetadataService metadataService;
  private GoogleIsbnImportService googleIsbnImportService;
  private SwisscoveryImportService swisscoveryImportService;
  private MessageService messageService;
  private DuplicateService duplicateService;

  private boolean checkForDuplicates;

  public MetadataImportWorkFlowService(DoiCrossrefImportService doiCrossrefImportService, ArxivImportService arxivImportService,
          PmidImportService pmidImportService, InspireHepImportService inspireHepImportService,
          EuropePmcImportService europePmcImportService, DataCiteImportService dataCiteImportService,
          UnpaywallImportService unpaywallImportService, MetadataService metadataService,
          GoogleIsbnImportService googleIsbnImportService, SwisscoveryImportService swisscoveryImportService,
          MessageService messageService, AouProperties aouProperties, DuplicateService duplicateService) {
    this.doiCrossrefImportService = doiCrossrefImportService;
    this.arxivImportService = arxivImportService;
    this.pmidImportService = pmidImportService;
    this.inspireHepImportService = inspireHepImportService;
    this.europePmcImportService = europePmcImportService;
    this.dataCiteImportService = dataCiteImportService;
    this.unpaywallImportService = unpaywallImportService;
    this.metadataService = metadataService;
    this.googleIsbnImportService = googleIsbnImportService;
    this.swisscoveryImportService = swisscoveryImportService;
    this.messageService = messageService;
    this.duplicateService = duplicateService;
    this.checkForDuplicates = aouProperties.getMetadata().getSearch().getDuplicates().isCheckForDuplicates();

  }

  public String getMetadataFromDoi(String doi, boolean isXmlFormat) {
    if (!ValidationTool.isValidDOI(doi)) {
      throw new SolidifyRuntimeException(this.messageService.get("deposit.error.identifiers.doi.invalid", new Object[] { doi }));
    }

    this.checkForDuplicate(Publication.ImportSource.DOI, doi);

    DepositDoc depositDoc = new DepositDoc();
    try {
      depositDoc = this.inspireHepImportService.searchInspireHep(doi, false, depositDoc);
      //complete metadata in crossref when it has been imported by InspireHep
      depositDoc = this.doiCrossrefImportService.searchDoi(doi, depositDoc);

    } catch (SolidifyRuntimeException e) {
      try {
        depositDoc = this.doiCrossrefImportService.searchDoi(doi, depositDoc);
      } catch (SolidifyRuntimeException exception) {
        try {
          depositDoc = this.dataCiteImportService.searchDoi(doi, depositDoc);
        } catch (SolidifyRuntimeException ex) {
          throw new SolidifyHttpErrorException(HttpStatus.NOT_FOUND,
                  this.messageService.get("deposit.error.identifiers.doi.not_found", new Object[] { doi }));
        }
      }
    }

    try {
      depositDoc = this.pmidImportService.searchDoi(doi, depositDoc);
      String pmcId = depositDoc.getIdentifiers().getPmcid();
      if (!StringTool.isNullOrEmpty(pmcId)) {
        depositDoc = this.europePmcImportService.searchEuropePmcId(pmcId, depositDoc);
      }
    } catch (SolidifyRuntimeException e) {
      log.warn("Unable to import from pmid with the DOI '{}'", doi);
    }

    if (isXmlFormat) {
      return this.doiCrossrefImportService.transformDepositDocToMetadata(depositDoc);
    } else {
      return this.doiCrossrefImportService.transformDepositDocToFormData(depositDoc);
    }
  }

  public String getMetadataFromPmid(String pmid, boolean isXmlFormat) {
    if (!ValidationTool.isValidPMID(pmid)) {
      throw new SolidifyRuntimeException(this.messageService.get("deposit.error.identifiers.pmid.invalid", new Object[] { pmid }));
    }

    this.checkForDuplicate(Publication.ImportSource.PMID, pmid);

    DepositDoc depositDoc = new DepositDoc();
    depositDoc = this.pmidImportService.searchPmid(pmid, depositDoc);
    String pmcId = depositDoc.getIdentifiers().getPmcid();
    if (!StringTool.isNullOrEmpty(pmcId)) {
      depositDoc = this.europePmcImportService.searchEuropePmcId(pmcId, depositDoc);
    }

    if (isXmlFormat) {
      return this.pmidImportService.transformDepositDocToMetadata(depositDoc);
    } else {
      return this.pmidImportService.transformDepositDocToFormData(depositDoc);
    }
  }

  public String getMetadataFromArxiv(String arxivId, boolean isXmlFormat) {
    if (!ValidationTool.isValidArxivId(arxivId)) {
      throw new SolidifyRuntimeException(this.messageService.get("deposit.error.identifiers.arxiv.invalid", new Object[] { arxivId }));
    }

    this.checkForDuplicate(Publication.ImportSource.ARXIV, arxivId);

    DepositDoc depositDoc = new DepositDoc();
    try {
      depositDoc = this.inspireHepImportService.searchInspireHep(arxivId, true, depositDoc);
    } catch (SolidifyRuntimeException e) {
      try {
        depositDoc = this.arxivImportService.searchArxiv(arxivId, depositDoc);
      } catch (SolidifyRuntimeException ex) {
        throw new SolidifyHttpErrorException(HttpStatus.NOT_FOUND,
                this.messageService.get("deposit.error.identifiers.arxiv.not_found", new Object[] { arxivId }));
      }
    }
    if (isXmlFormat) {
      return this.inspireHepImportService.transformDepositDocToMetadata(depositDoc);
    } else {
      return this.inspireHepImportService.transformDepositDocToFormData(depositDoc);
    }
  }

  public void importFilesFromServices(Publication publication) throws IOException, NoSuchAlgorithmException {
    //import from arxiv if needed
    if (publication.getImportSource().equals(Publication.ImportSource.ARXIV)) {
      try {
        this.arxivImportService.createDocumentFileFromArxivMetadata(publication);
      } catch (SolidifyRuntimeException e) {
        this.inspireHepImportService.createDocumentFileFromInspireHepMetadata(publication);
      }
    } else if (publication.getImportSource().equals(Publication.ImportSource.DOI)) {
      this.doiCrossrefImportService.createDocumentFileFromCrossrefMetadata(publication);
      this.unpaywallImportService.createDocumentFileFromUnpaywall(publication);
    } else if (publication.getImportSource().equals(Publication.ImportSource.PMID)) {
      //check if there is pmcid, if so import files from (Europe)PMC
      String pmcId = this.getPmcIdFromPublicationMetadata(publication);
      if (!StringTool.isNullOrEmpty(pmcId)) {
        this.europePmcImportService.createDocumentFileFromEuropePmc(publication);
      }
      this.unpaywallImportService.createDocumentFileFromUnpaywall(publication);
    } else if (publication.getImportSource().equals(Publication.ImportSource.ISBN)) {
      this.unpaywallImportService.createDocumentFileFromUnpaywall(publication);
    }
  }

  private String getPmcIdFromPublicationMetadata(Publication publication) {
    final MetadataExtractor metadataExtractor = this.metadataService.getMetadataExtractor(publication.getMetadataVersion());
    return metadataExtractor.getPmcIdFromPublication(publication);
  }

  public String getMetadataFromIsbn(String isbn, boolean isXmlFormat) {
    if (!ValidationTool.isValidISBN(isbn)) {
      throw new SolidifyRuntimeException(this.messageService.get("deposit.error.identifiers.isbn.invalid", new Object[] { isbn }));
    }

    this.checkForDuplicate(Publication.ImportSource.ISBN, isbn);

    DepositDoc depositDoc = new DepositDoc();
    try {
      depositDoc = this.googleIsbnImportService.searchInGoogle(isbn, depositDoc);
    } catch (SolidifyRuntimeException e) {
      try {
        depositDoc = this.swisscoveryImportService.searchIsbn(isbn, depositDoc);
      } catch (SolidifyRuntimeException ex) {
        throw new SolidifyHttpErrorException(HttpStatus.NOT_FOUND,
                this.messageService.get("deposit.error.identifiers.isbn.not_found", new Object[] { isbn }));
      }
    }
    if (isXmlFormat) {
      return this.googleIsbnImportService.transformDepositDocToMetadata(depositDoc);
    } else {
      return this.googleIsbnImportService.transformDepositDocToFormData(depositDoc);
    }
  }

  private void checkForDuplicate(Publication.ImportSource source, String value) {
    if (this.checkForDuplicates) {
      ExistingPublicationInfo existingPublicationInfo = null;
      String message = null;
      switch (source) {
        case DOI:
          existingPublicationInfo = this.duplicateService.getExistingPublicationInfoByDoi(value);
          message = this.messageService.get("deposit.error.identifiers.doi.already_exists", new Object[] { value });
          break;
        case PMID:
          existingPublicationInfo = this.duplicateService.getExistingPublicationInfoByPmid(value);
          message = this.messageService.get("deposit.error.identifiers.pmid.already_exists", new Object[] { value });
          break;
        case ARXIV:
          existingPublicationInfo = this.duplicateService.getExistingPublicationInfoByArxivId(value);
          message = this.messageService.get("deposit.error.identifiers.arxiv.already_exists", new Object[] { value });
          break;
        case ISBN:
          existingPublicationInfo = this.duplicateService.getExistingPublicationInfoByIsbn(value);
          message = this.messageService.get("deposit.error.identifiers.isbn.already_exists", new Object[] { value });
          break;
      }
      if (existingPublicationInfo != null) {
        final Map<Object, Object> params = new HashMap<>();
        if (!StringTool.isNullOrEmpty(existingPublicationInfo.getUrl())) {
          params.put("url", existingPublicationInfo.getUrl());
        } else {
          params.put("resId", existingPublicationInfo.getResId());
        }
        throw new AouRuntimeException(HttpStatus.BAD_REQUEST, message, ErrorCodes.IDENTIFIER_ALREADY_EXISTS.value(), params);
      }
    }
  }

}
