package ch.unige.aou.service.schedule;

import java.util.ArrayList;
import java.util.List;
import java.util.Optional;
import java.util.concurrent.ScheduledFuture;

import org.springframework.scheduling.concurrent.ThreadPoolTaskScheduler;
import org.springframework.scheduling.support.CronTrigger;
import org.springframework.stereotype.Service;

import ch.unige.solidify.exception.SolidifyRuntimeException;
import ch.unige.solidify.service.MessageService;

import ch.unige.aou.config.AouProperties;
import ch.unige.aou.model.schedule.ScheduledTask;
import ch.unige.aou.service.AouService;
import ch.unige.aou.task.AbstractTaskRunnable;

@Service
public class SchedulerService extends AouService {

  public enum SchedulingCancelType {INTERRUPT_RUNNING_TASK, DONT_INTERRUPT_RUNNING_TASK}

  private ThreadPoolTaskScheduler taskScheduler;

  /**
   * List containing the scheduled tasks runnables (= the tasks implementations)
   * <p>
   * Each AbstractTaskRunnable contains a reference to its corresponding ScheduledTask and ScheduledFuture
   */
  private List<AbstractTaskRunnable> scheduledTasks = new ArrayList<>();

  private int additionalThreadsSize;

  public SchedulerService(AouProperties aouProperties, MessageService messageService, ThreadPoolTaskScheduler taskScheduler) {
    super(messageService);
    this.taskScheduler = taskScheduler;
    this.additionalThreadsSize = aouProperties.getParameters().getSchedulerThreadPoolAdditionalSize();
  }

  /****************************************************************/

  /**
   * Check if a task is already scheduled
   *
   * @param scheduledTask
   * @return
   */
  public boolean isTaskScheduled(ScheduledTask scheduledTask) {
    return this.getScheduledAbstractTaskRunnable(scheduledTask).isPresent();
  }

  /**
   * Cancel the scheduling of a task
   *
   * @param scheduledTask
   * @param killRunningTask If the task is running when the method is called, indicates if its execution must be stopped or if it can goes to the end
   */
  public void cancelTaskScheduling(ScheduledTask scheduledTask, SchedulingCancelType schedulingCancelType) {
    Optional<AbstractTaskRunnable> abstractTaskRunnableOptional = this.getScheduledAbstractTaskRunnable(scheduledTask);
    if (abstractTaskRunnableOptional.isPresent()) {
      ScheduledFuture<AbstractTaskRunnable> scheduledFuture = abstractTaskRunnableOptional.get().getScheduledFuture();
      if (!scheduledFuture.isDone() && !scheduledFuture.isCancelled()) {
        boolean interruptRunningTask = schedulingCancelType == SchedulingCancelType.INTERRUPT_RUNNING_TASK;
        if (scheduledFuture.cancel(interruptRunningTask)) {
          this.removeScheduledAbstractTaskRunnable(scheduledTask);
        } else {
          throw new SolidifyRuntimeException("unable to cancel task '" + scheduledTask.getResId() + "' scheduling");
        }
      } else {
        throw new SolidifyRuntimeException("task '" + scheduledTask.getResId() + "' is already done or cancelled");
      }
    } else {
      throw new SolidifyRuntimeException("task '" + scheduledTask.getResId() + "' is not scheduled");
    }
  }

  /**
   * Start the scheduling of a task.
   * <p>
   * It uses a CronTrigger that ensure the same task is not started before its last execution is over.
   * Different tasks can run in parallel as the the ThreadPoolTaskScheduler pool size is automatically resized according to the number
   * of scheduled tasks.
   *
   * @param runnableTask
   */
  public void startTaskScheduling(AbstractTaskRunnable runnableTask) {
    if (Boolean.TRUE.equals(runnableTask.getScheduledTask().isEnabled())) {

      // scheduled a task by using a cron expression
      // Note: If a task run is not over when the next run should start, the CronTrigger skips this next run
      ScheduledFuture<AbstractTaskRunnable> scheduledFuture = (ScheduledFuture<AbstractTaskRunnable>) this.taskScheduler
              .schedule(runnableTask, new CronTrigger(runnableTask.getScheduledTask().getCronExpression()));
      runnableTask.setScheduledFuture(scheduledFuture);

      this.scheduledTasks.add(runnableTask);

      // set the scheduler pool size to allow different scheduled tasks to run in parallel.
      // Add some margin to allow applications's @Scheduled methods to run in parallel as well
      this.taskScheduler.setPoolSize(this.scheduledTasks.size() + this.additionalThreadsSize);
    } else {
      throw new SolidifyRuntimeException("A disabled task cannot be scheduled ('" + runnableTask.getScheduledTask().getResId() + "')");
    }
  }

  /**
   * Cancel all scheduled tasks started by this service
   */
  public void cancelAllTasksScheduling() {
    for (AbstractTaskRunnable abstractTaskRunnable : this.scheduledTasks) {
      ScheduledFuture<AbstractTaskRunnable> scheduledFuture = abstractTaskRunnable.getScheduledFuture();
      if (!scheduledFuture.isDone() && !scheduledFuture.isCancelled()) {
        scheduledFuture.cancel(false);
      }
    }
    this.scheduledTasks.clear();
  }

  /****************************************************************/

  private Optional<AbstractTaskRunnable> getScheduledAbstractTaskRunnable(ScheduledTask scheduledTask) {
    return this.scheduledTasks.stream().filter(taskRunnable -> taskRunnable.getScheduledTask().getResId().equals(scheduledTask.getResId()))
            .findFirst();
  }

  private boolean removeScheduledAbstractTaskRunnable(ScheduledTask scheduledTask) {
    Optional<AbstractTaskRunnable> abstractTaskRunnableOptional = this.getScheduledAbstractTaskRunnable(scheduledTask);
    if (abstractTaskRunnableOptional.isPresent()) {
      return this.scheduledTasks.remove(abstractTaskRunnableOptional.get());
    }
    return false;
  }
}
