package ch.unige.aou.service;

import java.util.NoSuchElementException;

import org.springframework.stereotype.Service;

import ch.unige.aou.business.CommentService;
import ch.unige.aou.business.PersonService;
import ch.unige.aou.business.PublicationService;
import ch.unige.aou.controller.AouControllerAction;
import ch.unige.aou.model.publication.Comment;
import ch.unige.aou.model.publication.Publication;

@Service("commentPermissionService")
public class CommentPermissionService extends AbstractPermissionService {

  private CommentService commentService;
  private PublicationService publicationService;

  public CommentPermissionService(PersonService personService, CommentService commentService, PublicationService publicationService) {
    super(personService);
    this.commentService = commentService;
    this.publicationService = publicationService;
  }

  public boolean isAllowed(String targetResId, String actionString) {
    if (this.canBypassPermissionCheck()) {
      return true;
    }

    Comment existingResource = this.getExistingResource(targetResId);
    if (existingResource == null) {
      throw new NoSuchElementException(targetResId);
    }

    String personId = this.getPersonAuthenticatedId();
    AouControllerAction action = this.getControllerAction(actionString);
    return this.isAllowedToPerformActionOnResource(personId, existingResource, null, action);
  }

  /******************************************************************************************************/

  private Comment getExistingResource(String resId) {
    return this.commentService.findOne(resId);
  }

  /******************************************************************************************************/

  private boolean isAllowedToPerformActionOnResource(String personId, Comment existingResource, Comment updateResource,
          AouControllerAction action) {

    final Comment resourceToCheck = (Comment) this.getResourceToCheckPermissionOn(existingResource, updateResource);

    // the only action checked by CommentPermissionService so far is to get a comment
    if (action != AouControllerAction.GET_DEPOSIT_COMMENT) {
      return false;
    }

    Comment comment = this.commentService.findOne(resourceToCheck.getResId());
    Publication publication = this.publicationService.findOne(comment.getPublicationId());

    return this.userIsAllowedToReadComment(personId, comment, publication);
  }

  /**
   * Return true if the person can read the comment.
   * It is the case if:
   * 1) comment only for validators:
   * - the person is validator of the publication
   * <p>
   * 2) comment is not reserved to validators:
   * - person is creator of the publication
   * - person is contributor of the publication
   * - person is validator of the publication
   * - person has written the comment
   *
   * @param personId
   * @param comment
   * @param publication
   * @return
   */
  private boolean userIsAllowedToReadComment(String personId, Comment comment, Publication publication) {

    if (comment.isOnlyForValidators() && this.personService.isAllowedToValidate(personId, publication)) {
      return true;
    } else {
      if (publication.getCreator().getResId().equals(personId) // user is publication's depositor
              || this.publicationService.isPersonPublicationContributor(personId, publication.getResId()) // user is a contributor
              || this.personService.isAllowedToValidate(personId, publication) // user is validator
              || comment.getCreatorId().equals(personId) // user is comment's author
      ) {
        return true;
      }
    }

    return false;
  }
}
